﻿using C1.C1Schedule;
using Google.Apis.Calendar.v3.Data;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;

namespace Scheduler.GoogleCalendarSyncer
{
    /// <summary>
    /// Class used for converting Scheduler Appointments to Google Events and vice versa.
    /// </summary>
    public static class Mapper
    {
        /// <summary>
        /// Converts event to appointment.
        /// </summary>
        public static readonly Expression<Func<Event, Appointment>> AsAppointment = source => new Appointment()
        {
            Subject = source.Summary,
            Body = source.Description,
            Location = source.Location,
            Label = new Label() { Color = GoogleCalendarColors.FirstOrDefault(x => x.Id == source.ColorId).Color },
            Start = source.Start.DateTime == null ? DateTime.Parse(source.Start.Date) : source.Start.DateTime.Value,
            End = source.End.DateTime == null ? DateTime.Parse(source.End.Date) : source.End.DateTime.Value
        };

        /// <summary>
        /// Converts event to appointment.
        /// </summary>
        public static readonly Func<Event, Appointment> ConvertToAppointment = AsAppointment.Compile();

        /// <summary>
        /// Converts appointment to event.
        /// </summary>
        public static readonly Expression<Func<Appointment, Event>> AsEvent = source => new Event()
        {
            Summary = source.Subject,
            Description = source.Body,
            Location = source.Location,
            ColorId = source.Label != null ? GoogleCalendarColors.FirstOrDefault(x => x.Color == source.Label.Color).Id : null,
            Start = new EventDateTime() { DateTime = source.Start },
            End = new EventDateTime() { DateTime = source.End }
        };

        /// <summary>
        /// Converts appointment to event.
        /// </summary>
        public static readonly Func<Appointment, Event> ConvertToEvent = AsEvent.Compile();

        /// <summary>
        /// Creates a label with color based on color id.
        /// </summary>
        /// <param name="googleColorId">
        ///  Id of the color. 
        ///  Ranges from 1 to 11
        /// </param>
        /// <returns></returns>
        public static Label GetAppointmentLabelFromColorId(string googleColorId)
        {
            if (!string.IsNullOrEmpty(googleColorId) && GoogleCalendarColors.Select(x => x.Id).Contains(googleColorId))
            {
                var calendarColor = GoogleCalendarColors.FirstOrDefault(x => x.Id == googleColorId);
                return new Label() { Color = calendarColor.Color, Text = calendarColor.ColorName };
            }

            return new Label() { Color = GoogleCalendarColors.First().Color, Text = GoogleCalendarColors.First().ColorName };
        }

        /// <summary>
        /// Google calendar colors collection. (Google calendar only supports colors present in this collection)
        /// </summary>
        public static readonly List<GoogleCalendarColor> GoogleCalendarColors = new List<GoogleCalendarColor>
        {
            new GoogleCalendarColor(){ Id = null, HexColorCode = "#039be5", ColorName = "Peacock"},
            new GoogleCalendarColor(){ Id = "1", HexColorCode = "#7986cb", ColorName = "Lavender" },
            new GoogleCalendarColor(){ Id = "2", HexColorCode = "#33b679", ColorName = "Sage"},
            new GoogleCalendarColor(){ Id = "3", HexColorCode = "#8e24aa", ColorName = "Grape"},
            new GoogleCalendarColor(){ Id = "4", HexColorCode = "#e67c73", ColorName = "Flamingo"},
            new GoogleCalendarColor(){ Id = "5", HexColorCode = "#f6c026", ColorName = "Banana"},
            new GoogleCalendarColor(){ Id = "6", HexColorCode = "#f5511d", ColorName = "Tangerine"},
            new GoogleCalendarColor(){ Id = "7", HexColorCode = "#039be5", ColorName = "Peacock"},
            new GoogleCalendarColor(){ Id = "8", HexColorCode = "#616161", ColorName = "Graphite"},
            new GoogleCalendarColor(){ Id = "9", HexColorCode = "#3f51b5", ColorName = "Blueberry"},
            new GoogleCalendarColor(){ Id = "10", HexColorCode = "#0b8043", ColorName = "Basil"},
            new GoogleCalendarColor(){ Id = "11", HexColorCode = "#d60000", ColorName = "Tomato"}
        };
    }
}
