﻿using C1.C1Schedule;
using Google.Apis.Auth.OAuth2;
using Google.Apis.Calendar.v3;
using Google.Apis.Calendar.v3.Data;
using Google.Apis.Services;
using Google.Apis.Util.Store;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

namespace Scheduler.GoogleCalendarSyncer
{

    public class GoogleEventAppointmentAdapter : IAppointmentAdapter
    {
        private static Dictionary<Appointment, string> _eventIdMapping;
        private CalendarService _service;
        private bool _manualSync;
        //calendarId is your primary gmail account associated with the calendar.
        private string _calendarId;
        private IList<Event> _fetchedEvents;
        // If syncing is manual then added appointment will be stored in this list
        private List<Appointment> _addedAppointments;
        // If syncing is manual then deleted appointment will be stored in this list
        private List<Appointment> _deletedAppointments;
        /// <summary>
        /// Fires before starting of syncing operation.
        /// </summary>
        public event EventHandler<SyncEventArgs> SyncStarting;
        /// <summary>
        /// Fires when syncing is finished.
        /// </summary>
        public event EventHandler SyncFinished;
        /// <summary>
        /// Fires when syncing fails.
        /// </summary>
        public event EventHandler SyncFailed;

        /// <summary>
        /// Gets or sets whether appointments are manually synced with Google Calendar.
        /// </summary>
        public bool ManualSync
        {
            get
            {
                return _manualSync;
            }
            set
            {
                if (value)
                {
                    if (_addedAppointments == null)
                        _addedAppointments = new List<Appointment>();

                    if (_deletedAppointments == null)
                        _deletedAppointments = new List<Appointment>();
                }
                else
                {
                    _addedAppointments = null;
                    _deletedAppointments = null;
                }

                _manualSync = value;
            }
        }

        /// <summary>
        /// Gets whether there are any added or deleted apoointment needs to be synced.
        /// </summary>
        public bool CanSync
        {
            get
            {
                return _addedAppointments.IsNullOrEmpty() && _deletedAppointments.IsNullOrEmpty() ? false : true;
            }
        }

        #region
        /// <summary>
        ///     Gets events based on parameters criteria.
        /// </summary>
        /// <param name="requestParameters">    
        /// </param>
        /// <returns></returns>
        public AppointmentsResponse GetAppointments(RequestParameters requestParameters)
        {
            EventsResource.ListRequest request = _service.Events.List(_calendarId);
            request.TimeMin = requestParameters.MinDate;
            request.TimeMax = requestParameters.MaxDate;
            request.SingleEvents = requestParameters.SingleEvents;
            request.ShowDeleted = requestParameters.ShowDeleted;
            request.MaxResults = requestParameters.MaxResults;
            request.PageToken = requestParameters.PageToken;
            if (request.SingleEvents.Value)
            {
                request.OrderBy = requestParameters.OrderBy == OrderBy.StartTime ? EventsResource.ListRequest.OrderByEnum.StartTime : EventsResource.ListRequest.OrderByEnum.Updated;
            }
            var response = request.Execute();
            _fetchedEvents = response.Items;

            // Converting events to appointments
            var appointments = _fetchedEvents.AsQueryable().Select(Mapper.AsAppointment).ToList();

            // For mapping each appointment with event Id
            appointments.ForEach(a => MapAppointmentWithEventId(a, _fetchedEvents[appointments.IndexOf(a)].Id));

            return new AppointmentsResponse()
            {
                Appointments = appointments,
                NextPageToken = response.NextPageToken
            };
        }

        /// <summary>
        ///     Adds appointment to google calendar
        /// </summary>
        /// <param name="appointment"></param>
        public void AddAppointment(Appointment appointment)
        {
            if (!_manualSync)
            {
                Event addedEvent = _service.Events.Insert(Mapper.ConvertToEvent(appointment), _calendarId).Execute();
                MapAppointmentWithEventId(appointment, addedEvent.Id);
                _fetchedEvents.Add(addedEvent);
            }
            else
            {
                if (!_addedAppointments.Contains(appointment))
                    _addedAppointments.Add(appointment);
            }
        }

        /// <summary>
        ///     Deletes appointment from google calendar.
        /// </summary>
        /// <param name="appointment"></param>
        public void DeleteAppointment(Appointment appointment)
        {
            if (IsDeletedEvent(GetMappedEventId(appointment)))
                return;

            if (!_manualSync)
            {
                _service.Events.Delete(_calendarId, GetMappedEventId(appointment)).Execute();
                var deletedEvent = _fetchedEvents.FirstOrDefault(x => x.Id == GetMappedEventId(appointment));
                _fetchedEvents.Remove(deletedEvent);
                RemoveMapping(appointment);
            }
            else
            {
                if (_addedAppointments.Contains(appointment))
                    _addedAppointments.Remove(appointment);
                else
                    _deletedAppointments.Add(appointment);
            }
        }

        /// <summary>
        /// Updates an edited appointment to google calendar. This method only uses auto sync mode.
        /// </summary>
        /// <param name="appointment"></param>
        public void UpdateAppointment(Appointment appointment)
        {
            if (_manualSync)
                return;

            if (IsDeletedEvent(GetMappedEventId(appointment)))
                return;

            var editedEvent = Mapper.ConvertToEvent(appointment);
            _service.Events.Update(editedEvent, _calendarId, GetMappedEventId(appointment)).Execute();
        }

        /// <summary>
        ///     Syncs all the added and deleted appointments with google calendar.
        /// </summary>
        public async Task Sync()
        {
            if (!CanSync)
                return;

            try
            {
                // Starting Sync
                SyncEventArgs syncArgs = new SyncEventArgs(_addedAppointments, _deletedAppointments);
                SyncStarting?.Invoke(this, syncArgs);

                if (syncArgs.Cancel)
                    return;

                if (!_addedAppointments.IsNullOrEmpty())
                {
                    foreach (var appointment in syncArgs.AddedAppointments)
                    {
                        var addedEvent = await _service.Events.Insert(Mapper.ConvertToEvent(appointment), _calendarId).ExecuteAsync();
                        MapAppointmentWithEventId(appointment, addedEvent.Id);
                        _fetchedEvents.Add(addedEvent);
                    }
                    _addedAppointments.Clear();
                }

                if (!_deletedAppointments.IsNullOrEmpty())
                {
                    foreach (var appointment in syncArgs.DeletedAppointments)
                    {
                        await _service.Events.Delete(_calendarId, GetMappedEventId(appointment)).ExecuteAsync();
                        var deletedEvent = _fetchedEvents.FirstOrDefault(x => x.Id == GetMappedEventId(appointment));
                        _fetchedEvents.Remove(deletedEvent);
                        RemoveMapping(appointment);
                    }
                    _deletedAppointments.Clear();
                }

                // Syncing Finished
                SyncFinished?.Invoke(this, new EventArgs());
            }
            catch
            {
                // Sync Failed
                SyncFailed?.Invoke(this, new EventArgs());
            }
        }

        #endregion

        #region Private Methods

        /// <summary>
        /// Checks whether the event is already deleted.
        /// </summary>
        /// <param name="eventId"></param>
        /// <returns></returns>
        private bool IsDeletedEvent(string eventId)
        {
            return _fetchedEvents.FirstOrDefault(e => e.Id == eventId).Status == "cancelled";
        }

        /// <summary>
        /// Maps event id with appointment
        /// </summary>
        /// <param name="appointment"></param>
        /// <param name="eventId"></param>
        private void MapAppointmentWithEventId(Appointment appointment, string eventId)
        {
            if (!_eventIdMapping.ContainsKey(appointment))
            {
                _eventIdMapping.Add(appointment, eventId);
            }
        }

        /// <summary>
        /// Gets mapped event id for appointment
        /// </summary>
        /// <param name="appointment"></param>
        /// <returns></returns>
        private string GetMappedEventId(Appointment appointment)
        {
            if (_eventIdMapping.ContainsKey(appointment))
            {
                return _eventIdMapping[appointment];
            }

            return null;
        }

        /// <summary>
        /// Removes mapping of event id.
        /// </summary>
        /// <param name="appointment"></param>
        private void RemoveMapping(Appointment appointment)
        {
            if (_eventIdMapping.ContainsKey(appointment))
            {
                _eventIdMapping.Remove(appointment);
            }
        }

        #endregion

        public GoogleEventAppointmentAdapter(CalendarService service,bool manualSync = false)
        {
            this._service = service;
            _eventIdMapping = new Dictionary<Appointment, string>();
            _calendarId = "primary";
            ManualSync = manualSync;
        }

    }


    /// <summary>
    ///  Class used for syncing of appointments between C1Scheduler and Google calendar.
    /// </summary>
    public class SyncManager
    {
        #region Fields
        private string _credentialFilePath;
        private string _accessTokenFilePath = Path.Combine("Credentials", "token.json");
        private string[] _scopes = { CalendarService.Scope.Calendar };
        private UserCredential _credential;
        private ClientSecrets _clientSecrets;
        private string _appName;
        private bool _manualSync;
        private bool _loggedIn;
        #endregion

        /// <summary>
        /// 
        /// </summary>
        public GoogleEventAppointmentAdapter Adapter
        {
            get;private set;
        }

        /// <summary>
        /// Fires when login process starts.
        /// </summary>
        public event EventHandler LoggingIn;
       /// <summary>
       ///  Fires when login process completes.
       /// </summary>
        public event EventHandler LoginSuccess;
        /// <summary>
        /// Fires when Login process is cancelled.
        /// </summary>
        public event EventHandler LoginFailed;

        #region Properties

        /// <summary>
        /// Location of credential.json file
        /// </summary>
        public string CredentialFilePath
        {
            get
            {
                return _credentialFilePath;
            }
            set
            {
                if (!string.IsNullOrEmpty(value))
                    _credentialFilePath = value;
            }
        }

        /// <summary>
        /// Gets whether the user have logged in or not.
        /// </summary>
        public bool LoggedIn
        {
            get
            {
                return _loggedIn;
            }
        }


        /// <summary>
        /// Client secrets generated using credential.json file.
        /// </summary>
        public ClientSecrets ClientSecrets
        {
            get
            {
                return _clientSecrets;
            }
            set
            {
                if(value!= null)
                    _clientSecrets = value;
            }
        }

        #endregion

        #region Public Methods

        /// <param name="appName"></param>
        ///     Application Name
        /// <param name="manualSync">
        ///     If it is false then adding or deleting appointments in scheduler will 
        ///     be immediately updated to Google Calendar.
        ///     
        ///     If it is true then adding or deleting appointments in scheduler
        ///     will not be synced immediately with Google Calendar. You have to
        ///     call the adapter's Sync method to sync all the added or deleted appointments.
        ///     (Default is false)
        /// </param>
        public SyncManager(string appName, bool manualSync = false)
        {
            _appName = appName;
            _loggedIn = false;
            _manualSync = manualSync;
        }

        /// <summary>
        ///    Authorizes the user and creates the calendar service. Call to this method will open the google login page.
        /// </summary>
        /// <param name="cancelAfterMilliseconds">
        ///  Auto cancel after specified milliseconds. Default is 30000 (30 seconds).
        /// </param>
        public async Task Login(int cancelAfterMilliseconds = 30000)
        {
            if (_clientSecrets != null)
            {
                await Login(_clientSecrets, cancelAfterMilliseconds);
            }
            else
            {
                using (var stream = new FileStream(_credentialFilePath, FileMode.Open, FileAccess.Read))
                {
                    await Login(GoogleClientSecrets.Load(stream).Secrets, cancelAfterMilliseconds);
                }
            }
        }

        /// <summary>
        ///   Authorizes the user and creates the calendar service. Call to this method will open the google login page.
        /// </summary>
        /// <param name="clientSecrets">
        ///  Use GoogleClientSecrets class to get your client secrets from credential.json
        /// </param>
        /// <param name="cancelAfterMilliseconds">
        ///  Auto cancel after specified milliseconds. Default is 30000 (30 seconds).
        /// </param>
        public async Task Login(ClientSecrets clientSecrets, int cancelAfterMilliseconds = 30000)
        {
            LoggingIn?.Invoke(this, new EventArgs());
            CancellationTokenSource tokenSource = new CancellationTokenSource();
            CancellationToken token = tokenSource.Token;
            tokenSource.CancelAfter(cancelAfterMilliseconds);

            // auto cancel delegate
            token.Register(new System.Action(() => {
                if(!_loggedIn)
                    LoginFailed?.Invoke(this, new EventArgs());
            }));

            // authorization
            _credential = await GoogleWebAuthorizationBroker.AuthorizeAsync(
                clientSecrets,
                _scopes,
                "user",
                token,
                new FileDataStore(_accessTokenFilePath, true));

            // Creating Google Calendar API service.
            var service = new CalendarService(new BaseClientService.Initializer()
            {
                HttpClientInitializer = _credential,
                ApplicationName = _appName,
            });

            this.Adapter = new GoogleEventAppointmentAdapter(service, _manualSync);
            _loggedIn = true;
            LoginSuccess?.Invoke(this, new EventArgs());
        }

        /// <summary>
        /// Deletes the current logged in user info.
        /// </summary>
        public void Logout()
        {
            if (!_loggedIn)
                return;
            this.Adapter = null;

            if (Directory.Exists(_accessTokenFilePath))
            {
                Directory.Delete(_accessTokenFilePath, true);
            }

            _loggedIn = false;
        }

        #endregion
    }
}
