﻿using C1.WPF.Schedule;
using System;
using System.Windows;
using Scheduler.GoogleCalendarSyncer;

namespace SchedulerWithGoogleCalendar
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private SyncManager _syncManager;
        private AppointmentsResponse _appointmentsResponse;

        /// <summary>
        ///     Whether the current response has remaining data on next page.
        /// </summary>
        public bool HasNextPage
        {
            get
            {
                if (_appointmentsResponse == null)
                    return false;

                return _appointmentsResponse.NextPageToken != null;
            }
        }

        public MainWindow()
        {
            InitializeComponent();
            AddGoogleCalendarColorsToScheduler();
            InitializeSyncer();
            DataContext = this;
            Closing += WindowClosing;
        }

        private void WindowClosing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            if (_syncManager.LoggedIn)
            {
                _syncManager.Logout();
            }
        }

        private void InitializeSyncer()
        {
            _syncManager = new SyncManager("Scheduler")
            {
                CredentialFilePath = "Credentials\\credentials.json"
            };


            filterToolbar.PropertyChanged += (s, e) =>
            {
                if (e.PropertyName == "ManualSync")
                {
                    _syncManager.Adapter.ManualSync = filterToolbar.ManualSync;
                }
            };

            _syncManager.LoggingIn += (s, e) => { loadingScreen.Show(true, 60); };
            _syncManager.LoginSuccess += (s, e) => 
            {
                _syncManager.Adapter.SyncStarting += SyncStarting;
                _syncManager.Adapter.SyncFinished += SyncFinished;
                _syncManager.Adapter.SyncFailed += SyncFailed;
                loadingScreen.Hide(); 
            };
            _syncManager.LoginFailed += (s, e) => {
                Application.Current.Dispatcher.BeginInvoke(new System.Action(() => { loadingScreen.Hide(); }));
             };
        }

        private void AddGoogleCalendarColorsToScheduler()
        {
            _scheduler.DataStorage.LabelStorage.Clear();

            foreach (var calendarColor in Mapper.GoogleCalendarColors)
            {
                _scheduler.DataStorage.LabelStorage.Labels.Add(Mapper.GetAppointmentLabelFromColorId(calendarColor.Id));
            }
        }

        #region Fetching, Adding, Deleting, Updating appointments
        /// <summary>
        /// Fetches appointments and adds them to scheduler.
        /// </summary>
        /// <param name="requestParameters"></param>
        private void FetchAppointments(RequestParameters requestParameters)
        {
            _scheduler.AppointmentAdded -= SchedulerAppointmentAdded;
            _scheduler.AppointmentDeleted -= SchedulerAppointmentDeleted;

            if (requestParameters.MinDate > requestParameters.MaxDate)
            {
                MessageBox.Show("Min date should be less than Max date", "Invalid date range", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            _scheduler.DataStorage.AppointmentStorage.Appointments.Clear();

            _appointmentsResponse = _syncManager.Adapter.GetAppointments(requestParameters);

            _scheduler.BeginUpdate();
            foreach (var appointment in _appointmentsResponse.Appointments)
            {
                _scheduler.DataStorage.AppointmentStorage.Appointments.Add(appointment);
            }
            _scheduler.EndUpdate();

            _scheduler.AppointmentAdded += SchedulerAppointmentAdded;
            _scheduler.AppointmentDeleted += SchedulerAppointmentDeleted;
        }

        private void SchedulerAppointmentAdded(object sender, AppointmentActionEventArgs e)
        {
            if (_syncManager.LoggedIn)
                _syncManager.Adapter.AddAppointment(e.Appointment);
        }

        private void SchedulerAppointmentDeleted(object sender, AppointmentActionEventArgs e)
        {
            if(_syncManager.LoggedIn)
                _syncManager.Adapter.DeleteAppointment(e.Appointment);
        }

        private void RefreshClicked(object sender, RoutedEventArgs e)
        {
            if(_syncManager.LoggedIn)
                FetchAppointments(filterToolbar.GetRequestParameters());
        }

        private void NextPageClicked(object sender, RoutedEventArgs e)
        {
            var parameters = filterToolbar.GetRequestParameters();
            parameters.PageToken = _appointmentsResponse.NextPageToken;
            FetchAppointments(parameters);
        }

        private void SchedulerBeforeAppointmentSave(object sender, AppointmentActionEventArgs e)
        {
            if (_scheduler.DataStorage.AppointmentStorage.Appointments.Contains(e.Appointment) && _syncManager.LoggedIn)
            {
                _syncManager.Adapter.UpdateAppointment(e.Appointment);
            }
        }
        #endregion

        #region Syncing
        private async void SyncClicked(object sender, RoutedEventArgs e)
        {
            if (!_syncManager.Adapter.CanSync)
            {
                MessageBox.Show("No appointments available to sync.", "No sync available", MessageBoxButton.OK, MessageBoxImage.Information);
            }
            else
            {
                await _syncManager.Adapter.Sync();
            }
        }

        private void SyncFailed(object sender, EventArgs e)
        {
            MessageBox.Show("Something went wrong. Sync failed.", "Sync Fialed", MessageBoxButton.OK, MessageBoxImage.Error);
        }

        private void SyncFinished(object sender, EventArgs e)
        {
            MessageBox.Show("Sync finished successfully.", "Success", MessageBoxButton.OK);
        }

        private void SyncStarting(object sender, SyncEventArgs e)
        {
            string message = e.AddedAppointments.Count > 0 ? $"{e.AddedAppointments.Count} added " : string.Empty;
            message += e.DeletedAppointments.Count > 0 ? " and " : string.Empty;
            message += e.DeletedAppointments.Count > 0 ? $"{e.DeletedAppointments.Count} deleted " : string.Empty;
            message += "appointments will be synced. Do you want to continue ?";

            var result = MessageBox.Show(message, "Continue", MessageBoxButton.YesNo, MessageBoxImage.Question, MessageBoxResult.Yes);

            if (result == MessageBoxResult.No)
                e.Cancel = true;
        }
        #endregion

        #region Login and Logout
        private async void LoginLogoutClick(object sender, RoutedEventArgs e)
        {
            if (!_syncManager.LoggedIn)
            {
                try
                {
                    await _syncManager.Login(60000);
                    this.Topmost = true;
                    FetchAppointments(filterToolbar.GetRequestParameters());
                    btnLoginLogout.Content = "Logout";
                }
                catch
                {

                }
            }
            else
            {
                _syncManager.Logout();
                btnLoginLogout.Content = "Login";
                _scheduler.AppointmentDeleted -= SchedulerAppointmentDeleted;
                _scheduler.AppointmentAdded -= SchedulerAppointmentAdded;
                _scheduler.DataStorage.AppointmentStorage.Appointments.Clear();
            }
        }
        #endregion
    }
}

